# Refs in comments in this file point to https://www.rabbitmq.com/resources/specs/amqp0-9-1.pdf
from __future__ import annotations

import asyncio
import logging
import platform
import struct
from functools import singledispatchmethod
from typing import Dict, FrozenSet, Optional, Tuple, Type, Union, cast

import pamqp.common
import pamqp.exceptions
import pamqp.frame
import pamqp.header

from . import _exceptions, _heartbeat
from ._channel import AcceptableFrame, Channel
from ._channel_id_store import ChannelIdStore
from ._consume_channel import ConsumeChannel
from ._exceptions import ConnectionClosed, ConnectionClosedByHeartbeatMonitor
from ._exchange_channel import ExchangeChannel
from ._frame_writer import FrameWriter
from ._rpc_channel import RpcChannel

logger = logging.getLogger(__name__)


def _safe_unmarshal_frame(
    buf: bytearray,
) -> Optional[
    Tuple[
        int,
        int,
        Union[
            pamqp.base.Frame,
            pamqp.body.ContentBody,
            pamqp.header.ContentHeader,
            pamqp.heartbeat.Heartbeat,
        ],
    ]
]:
    if len(buf) < 7:
        return None
    (size,) = struct.unpack("!I", buf[3:7])
    frame_size = size + 8  # 7 bytes for header, 1 byte for frame-end
    if len(buf) < frame_size:
        return None
    n_bytes, channel_id, frame = pamqp.frame.unmarshal(bytes(buf[:frame_size]))
    # pamqp unmarshals the ProtocolHeader; but we handle that one specially.
    assert not isinstance(frame, pamqp.header.ProtocolHeader)
    return n_bytes, channel_id, frame


class Protocol(asyncio.Protocol):
    def __init__(
        self, username: str = "guest", password: str = "guest", virtual_host: str = "/"
    ):
        self._username = username
        self._password = password
        self._virtual_host = virtual_host
        self._buffer = bytearray()
        # Negotiation frames, keyed by class. ref: 2.2.4
        self._lifecycle: Dict[Type[pamqp.base.Frame], pamqp.base.Frame] = {}
        self._channels: Dict[int, Channel] = {}
        self._closing_channels: Dict[int, Optional[pamqp.commands.Channel.Close]] = {}
        self._exchange_channels: Dict[str, ExchangeChannel] = {}
        self._heartbeat_monitor: Optional[_heartbeat.HeartbeatMonitor] = None
        self._heartbeat_sender: Optional[_heartbeat.HeartbeatSender] = None
        self.open: asyncio.Future[None] = asyncio.Future()
        self._crash_exc: Optional[Exception] = None
        self._close_frame: Optional[pamqp.commands.Connection.Close] = None
        self._closing: bool = False

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self._transport = cast(asyncio.WriteTransport, transport)
        self._transport.write(pamqp.header.ProtocolHeader().marshal())  # ref: 2.2.4
        self._frame_writer = FrameWriter(self._transport, 0)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        # ... we'll look to many places for exceptions that may supercede the
        # one asyncio gives us.
        if self._crash_exc:
            exc = self._crash_exc
        elif self._close_frame:
            exc = _exceptions.ConnectionClosedByServer(
                self._close_frame.reply_code or 0, self._close_frame.reply_text
            )

        if self._heartbeat_monitor is not None:
            self._heartbeat_monitor.cancel()

        if self._heartbeat_sender is not None:
            self._heartbeat_sender.cancel()

        logger.info("Connection lost: %r", exc)

        if self._channels:
            closed_exc = ConnectionClosed()
            for channel in list(self._channels.values()):
                channel.accept_connection_closed(closed_exc)
            self._channels.clear()
            self._closing_channels.clear()

        if self.open.done():
            if exc:
                self.closed.set_exception(exc)
            else:
                self.closed.set_result(None)
        else:
            assert exc is not None
            self.open.set_exception(exc)

    def data_received(self, data: bytes) -> None:
        if self._heartbeat_monitor is not None:
            self._heartbeat_monitor.reset()

        self._buffer.extend(data)
        self._handle_buffer_frames()

    def _crash(self, exc: Exception) -> None:
        self._crash_exc = exc
        if pamqp.commands.Connection.Tune not in self._lifecycle:
            # ref: 2.2.4
            # "There is no hand-shaking for errors on connections that are not
            # fully open. Following successful protocol header negotiation,
            # which is defined in detail later, and prior to sending or
            # receiving Open or Open-Ok, a peer that detects an error MUST close
            # the socket without sending any further data."
            logger.exception(
                "Disconnecting because of error during negotiation", exc_info=exc
            )
            self._transport.abort()
        else:
            logger.exception("Sending Connection.Close because of error", exc_info=exc)
            if (
                isinstance(exc, pamqp.exceptions.PAMQPException)
                and hasattr(exc, "name")
                and hasattr(exc, "value")
            ):
                self.send_close_if_allowed(reply_code=exc.value, reply_text=exc.name)  # type: ignore[attr-defined]
            elif isinstance(exc, NotImplementedError):
                self.send_close_if_allowed(reply_code=540, reply_text="NOT-IMPLEMENTED")
            else:
                self.send_close_if_allowed(reply_code=541, reply_text="INTERNAL-ERROR")
            # ... and we expect RabbitMQ to handshake the closure

    def _handle_buffer_frames(self) -> None:
        """Unmarshal and handle all whole frames on `self._buffer`.

        When this method returns, `self._buffer` will be empty or contain the
        start of an incomplete frame.

        Log and call `_crash()` if handling fails.
        """
        while True:
            try:
                frame_result = _safe_unmarshal_frame(self._buffer)
            except pamqp.exceptions.UnmarshalingException:
                logger.exception("Invalid frame from RabbitMQ server")
                self._crash(
                    _exceptions.BadBytesFromServer("Invalid frame from RabbitMQ server")
                )
                return

            if frame_result is None:
                break

            n_bytes, channel_id, frame = frame_result
            self._buffer = self._buffer[n_bytes:]
            self._handle_frame_or_crash(channel_id, frame)

    def _handle_frame_or_crash(
        self,
        channel_id: int,
        frame: Union[
            pamqp.base.Frame,
            pamqp.body.ContentBody,
            pamqp.header.ContentHeader,
            pamqp.heartbeat.Heartbeat,
        ],
    ) -> None:
        logger.debug("accept_frame: %r", frame)
        try:
            self.accept_frame(frame, channel_id)
        except (pamqp.exceptions.PAMQPException, NotImplementedError) as err:
            self._crash(err)
            return

    def _assert_channel_id_0(
        self, frame: Union[AcceptableFrame, pamqp.heartbeat.Heartbeat], channel_id: int
    ) -> None:
        # ref: 4.2.3
        #
        # "The channel number MUST be zero for all heartbeat frames, and for
        # method, header and body frames that refer to the Connection class. A
        # peer that receives a non-zero channel number for one of these frames
        # MUST signal a connection exception with reply code 503 (command
        # invalid)."
        if channel_id != 0:
            raise pamqp.exceptions.AMQPCommandInvalid(
                "Got channel_id=%d on frame %r" % (channel_id, frame)
            )

    def accept_frame(
        self, frame: Union[AcceptableFrame, pamqp.heartbeat.Heartbeat], channel_id: int
    ) -> None:
        """Handle a single frame from the server.

        Raise NotImplementedError if it is an unexpected or incorrect frame.
        """
        if self.open.done():
            return self._accept_frame_after_handshake(frame, channel_id)
        else:
            self._assert_channel_id_0(frame, channel_id)
            return self._accept_handshake_frame(frame)

    @singledispatchmethod
    def _accept_frame_after_handshake(self, frame, channel_id: int) -> None:
        channel = self._channels[channel_id]  # KeyError means bug here or on server
        channel.accept_frame(frame)

    @_accept_frame_after_handshake.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Connection.Close, channel_id: int) -> None:
        self._close_frame = frame
        # ref: 2.2.4
        #  One peer (client or server) ends the connection (Close).
        #  The other peer hand-shakes the connection end (Close-Ok).
        self._frame_writer.send_frame(pamqp.commands.Connection.CloseOk())
        self._transport.close()  # asyncio will flush buffered data asynchronously

    @_accept_frame_after_handshake.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Connection.CloseOk, channel_id: int) -> None:
        # ref: 2.2.4
        #  One peer (client or server) ends the connection (Close).
        #  The other peer hand-shakes the connection end (Close-Ok).
        self._transport.abort()  # calls self.connection_lost()

    @_accept_frame_after_handshake.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.heartbeat.Heartbeat, channel_id: int) -> None:
        self._assert_channel_id_0(frame, channel_id)
        # _all_ network traffic counts as heartbeat frames, so we don't need to
        # do anything here. (We handle network traffic in self.data_received().)

    @_accept_frame_after_handshake.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Channel.Close, channel_id: int) -> None:
        # We handle Channel.Close and Channel.Close-Ok because they can race:
        # the client and server can each send Close simultaneously, in which
        # case we want to

        # We always send close-ok...
        FrameWriter(self._transport, channel_id).send_frame(
            pamqp.commands.Channel.CloseOk()
        )

        if channel_id in self._closing_channels:
            # We sent a Close() ourselves; we're still waiting for server CloseOk()
            #
            # This frame from the server might be an exception; store it.
            self._closing_channels[channel_id] = frame
        elif channel_id in self._channels:
            # We never sent Close(), so this channel is done
            channel = self._channels.pop(channel_id)
            channel.accept_channel_closed_by_server(frame)
            self._channel_id_store.release(channel_id)
            if hasattr(channel, "exchange_name"):
                del self._exchange_channels[channel.exchange_name]
        else:
            logger.error(
                "RabbitMQ sent Close on nonexistent channel; responding ok anyway"
            )

    @_accept_frame_after_handshake.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Channel.CloseOk, channel_id: int) -> None:
        close_frame = self._closing_channels.pop(channel_id)
        channel = self._channels.pop(channel_id)
        self._channel_id_store.release(channel_id)
        if hasattr(channel, "exchange_name"):
            del self._exchange_channels[channel.exchange_name]
        if close_frame is None:
            channel.accept_channel_closed_by_us()
        else:
            channel.accept_channel_closed_by_server(close_frame)

    @singledispatchmethod
    def _accept_handshake_frame(self, frame) -> None:
        raise NotImplementedError(
            "Unexpected frame during negotiation: %r" % frame
        )  # pragma: no cover

    @_accept_handshake_frame.register  # type: ignore [no-redef]
    def _(self, frame: pamqp.heartbeat.Heartbeat) -> None:
        pass  # already handled

    @_accept_handshake_frame.register  # type: ignore [no-redef]
    def _(self, frame: pamqp.commands.Connection.Start) -> None:
        assert pamqp.commands.Connection.Start not in self._lifecycle
        self._lifecycle[pamqp.commands.Connection.Start] = frame
        assert "PLAIN" in frame.mechanisms
        self._frame_writer.send_frame(
            pamqp.commands.Connection.StartOk(
                client_properties={
                    "platform": "%s %s (%s build %s)"
                    % (
                        platform.python_implementation(),
                        platform.python_version(),
                        *platform.python_build(),
                    ),
                    "product": "carehare",
                    "capabilities": {
                        "authentication_failure_close": True,
                        "publisher_confirms": True,
                    },
                    "information": "See https://github.com/cjworkbench/carehare",
                },
                mechanism="PLAIN",
                response=f"\x00{self._username}\x00{self._password}",
            ),
        )

    @_accept_handshake_frame.register  # type: ignore [no-redef]
    def _(self, frame: pamqp.commands.Connection.Tune) -> None:
        assert pamqp.commands.Connection.Start in self._lifecycle
        assert pamqp.commands.Connection.Tune not in self._lifecycle
        self._lifecycle[pamqp.commands.Connection.Tune] = frame
        self._frame_writer.send_frames(
            [
                pamqp.commands.Connection.TuneOk(
                    channel_max=frame.channel_max,
                    frame_max=frame.frame_max,
                    heartbeat=frame.heartbeat,
                ),
                pamqp.commands.Connection.Open(virtual_host=self._virtual_host),
            ]
        )
        self._heartbeat_sender = _heartbeat.HeartbeatSender(
            self._transport, frame.heartbeat
        )
        self._channel_id_store: ChannelIdStore = ChannelIdStore(frame.channel_max)

    @_accept_handshake_frame.register  # type: ignore [no-redef]
    def _(self, frame: pamqp.commands.Connection.OpenOk) -> None:
        assert pamqp.commands.Connection.Tune in self._lifecycle
        assert pamqp.commands.Connection.OpenOk not in self._lifecycle
        self._heartbeat_monitor = _heartbeat.HeartbeatMonitor(
            cast(
                pamqp.commands.Connection.Tune,
                self._lifecycle[pamqp.commands.Connection.Tune],
            ).heartbeat,
            on_dead=self._handle_heartbeat_monitor_dead,
        )
        self.open.set_result(None)
        self.closed: asyncio.Future[None] = asyncio.Future()

    @_accept_handshake_frame.register  # type: ignore [no-redef]
    def _(self, frame: pamqp.commands.Connection.Close) -> None:
        self._close_frame = frame

        if pamqp.commands.Connection.Tune in self._lifecycle:
            self._frame_writer.send_frame(pamqp.commands.Connection.CloseOk())
            self._transport.close()
        else:
            # ref: 2.2.4
            # "There is no hand-shaking for errors on connections that are not fully open"
            self._transport.abort()

        # Let `connection_lost()` complete self.open|self.closed.

    def _handle_heartbeat_monitor_dead(self):
        self._crash_exc = ConnectionClosedByHeartbeatMonitor()
        self._transport.abort()

    def send_close_if_allowed(self, reply_code: int = 0, reply_text: str = "") -> None:
        if not self.open.done():
            raise RuntimeError("Cannot send Close frame before opening.")

        if not self._closing and not self.closed.done():
            self._closing = True
            self._frame_writer.send_frame(
                pamqp.commands.Connection.Close(
                    reply_code=reply_code,
                    reply_text=reply_text,
                    class_id=0,
                    method_id=0,
                ),
            )

    def rpc(
        self,
        frame: pamqp.base.Frame,
        ignored_frame_types: FrozenSet[type] = frozenset(),
    ) -> asyncio.Future[None]:
        """Call a method remotely, with no return value.

        This pattern is useful for, say, queue.declare and such. The steps:

            1. Allocate a channel ID.
            2. Send "Open channel" + [frame] + "Close channel".
            3. Listen for the server response. (A Close means, "Exception".
               A Close-Ok means, "Success".)
            4. Release the channel ID.

        We return the Future that is set by the server response.

        Raise pamqp.exceptions.Exception if the server responds with an error.

        Raise NotImplementedError if the server responds with a frame we don't
        expect. Pass `ignore_frame_types` to allow and ignore certain frames.
        """
        channel_id = self._channel_id_store.acquire()
        channel = RpcChannel(
            frame,
            frame_writer=FrameWriter(self._transport, channel_id),
            ignored_frame_types=ignored_frame_types,
        )
        self._channels[channel_id] = channel
        self._closing_channels[channel_id] = None  # RPC channels close during init
        return channel.done

    def acking_consumer(
        self, queue_name: str, *, prefetch_size: int, prefetch_count: int
    ) -> ConsumeChannel:
        def on_closing():
            self._closing_channels[channel_id] = None

        channel_id = self._channel_id_store.acquire()
        channel = ConsumeChannel(
            frame_writer=FrameWriter(
                self._transport, channel_id, on_closing=on_closing
            ),
            queue_name=queue_name,
            prefetch_size=prefetch_size,
            prefetch_count=prefetch_count,
        )
        self._channels[channel_id] = channel
        return channel

    def _get_exchange_channel(self, exchange_name: str) -> ExchangeChannel:
        if exchange_name not in self._exchange_channels:
            frame_max = cast(
                pamqp.commands.Connection.Tune,
                self._lifecycle[pamqp.commands.Connection.Tune],
            ).frame_max
            channel_id = self._channel_id_store.acquire()
            channel = ExchangeChannel(
                exchange_name,
                frame_writer=FrameWriter(self._transport, channel_id),
                frame_max=frame_max,
            )
            self._channels[channel_id] = channel
            self._exchange_channels[exchange_name] = channel
        return self._exchange_channels[exchange_name]

    def publish(
        self, message: bytes, *, exchange_name: str, routing_key: str
    ) -> asyncio.Future[None]:
        exchange_channel = self._get_exchange_channel(exchange_name)
        return exchange_channel.publish(message, routing_key=routing_key)



from __future__ import annotations

import asyncio
from functools import singledispatchmethod
from typing import Dict, List

import pamqp.base
import pamqp.body
import pamqp.commands
import pamqp.exceptions
import pamqp.header

from ._channel import Channel
from ._exceptions import ChannelClosedByServer, ConnectionClosed, ServerSentNack
from ._frame_writer import FrameWriter


class ExchangeChannel(Channel):
    def __init__(
        self, exchange_name: str, *, frame_writer: FrameWriter, frame_max: int
    ) -> None:
        self.exchange_name = exchange_name
        self._frame_writer = frame_writer
        self._frame_max = frame_max
        self._delivering: Dict[int, asyncio.Future] = {}
        self._next_delivery_tag = 1

        self._frame_writer.send_frames(
            [
                pamqp.commands.Channel.Open(),
                pamqp.commands.Confirm.Select(nowait=True),
            ]
        )

    @singledispatchmethod
    def accept_frame(self, frame) -> None:
        raise NotImplementedError("Unexpected frame: %r" % frame)  # pragma: no cover

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Channel.OpenOk) -> None:
        pass

    def _find_delivery_tags(self, delivery_tag: int, multiple: bool) -> List[int]:
        if multiple:
            return [tag for tag in self._delivering.keys() if tag <= delivery_tag]
        else:
            return [delivery_tag]

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Basic.Ack) -> None:
        for tag in self._find_delivery_tags(frame.delivery_tag, frame.multiple):
            self._delivering.pop(tag).set_result(None)

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Basic.Nack) -> None:
        for tag in self._find_delivery_tags(frame.delivery_tag, frame.multiple):
            self._delivering.pop(tag).set_exception(ServerSentNack())

    def accept_channel_closed_by_server(
        self, frame: pamqp.commands.Channel.Close
    ) -> None:
        exc = ChannelClosedByServer(frame.reply_code or 0, frame.reply_text)
        for delivery in self._delivering.values():
            delivery.set_exception(exc)
        self._delivering.clear()

    def accept_channel_closed_by_us(self) -> None:
        raise NotImplementedError(
            "We never send Channel.Close; how did we get here?"
        )  # pragma: no cover

    def accept_connection_closed(self, exc: ConnectionClosed) -> None:
        for delivery in self._delivering.values():
            delivery.set_exception(exc)
        self._delivering.clear()

    def publish(self, message: bytes, routing_key: str = "") -> asyncio.Future[None]:
        body_frames = [
            pamqp.body.ContentBody(value=message[i : i + self._frame_max])
            for i in range(0, len(message), self._frame_max)
        ]

        self._frame_writer.send_frames(
            [
                pamqp.commands.Basic.Publish(
                    exchange=self.exchange_name, routing_key=routing_key
                ),
                pamqp.header.ContentHeader(body_size=len(message)),
                *body_frames,
            ]
        )
        done: asyncio.Future[None] = asyncio.Future()
        self._delivering[self._next_delivery_tag] = done
        self._next_delivery_tag += 1
        return done



from typing import Protocol, Union

import pamqp.base
import pamqp.body
import pamqp.header

from ._exceptions import ConnectionClosed

AcceptableFrame = Union[
    pamqp.base.Frame, pamqp.body.ContentBody, pamqp.header.ContentHeader
]


class Channel(Protocol):
    def accept_frame(self, frame: AcceptableFrame) -> None:
        """Handle a frame from the server.

        You can implement this as a @functools.singledispatchmethod.

        This method will never receive a `pamqp.commands.Channel.Close` or
        `pamqp.commands.Channel.CloseOk`. These are handled by the protocol,
        to handle races pertaining to closing on both ends.
        """
        raise NotImplementedError  # pragma: no cover

    def accept_channel_closed_by_server(self, frame: pamqp.commands.Channel.Close):
        """Handle the fact that this channel is shutting down normally.

        If implementor methods have returned any Futures, this method must
        complete those Futures.

        This is guaranteed to be the last method called by the protocol, and
        it's guaranteed not to be called twice.
        """

    def accept_channel_closed_by_us(self) -> None:
        """Handle the fact that we closed this channel.

        This will only occur if we sent a `Channel.Close` frame ourselves.

        This is guaranteed to be the last method called by the protocol, and
        it's guaranteed not to be called twice.
        """

    def accept_connection_closed(self, exc: ConnectionClosed) -> None:
        """Handle the fact that we're no longer connected to the server.

        If implementor methods have returned any Futures, this method must
        complete those Futures.

        This is guaranteed to be the last method called by the protocol, and
        it's guaranteed not to be called twice.
        """



import asyncio
import logging
from typing import Callable, List, Optional, Union

import pamqp.body
import pamqp.frame
import pamqp.header

logger = logging.getLogger(__name__)


class FrameWriter:
    def __init__(
        self,
        transport: asyncio.WriteTransport,
        channel_id: int,
        on_closing: Optional[Callable[[], None]] = None,
    ) -> None:
        self._transport = transport
        self._channel_id = channel_id
        self._on_closing = on_closing
        self._closing: bool = False

    def send_close(self):
        if self._closing:
            return

        self.send_frame(
            pamqp.commands.Channel.Close(
                reply_code=0, reply_text="", class_id=0, method_id=0
            )
        )
        self._closing = True
        if self._on_closing is not None:
            self._on_closing()

    def send_frame(
        self,
        frame: Union[
            pamqp.base.Frame, pamqp.body.ContentBody, pamqp.header.ContentHeader
        ],
    ) -> None:
        """Write a single frame to the transport, tagged with our channel_id

        Does not back-pressure.
        """
        if self._closing:
            logger.debug("not sending frame because we are closing: %r", frame)
            return

        logger.debug("send_frame: %r", frame)
        self._transport.write(pamqp.frame.marshal(frame, self._channel_id))

    def send_frames(
        self,
        frames: List[
            Union[pamqp.base.Frame, pamqp.body.ContentBody, pamqp.header.ContentHeader]
        ],
    ) -> None:
        """Write frames to the transport, tagged with our channel_id

        Does not back-pressure.
        """
        logger.debug("send_frames: %r", frames)
        self._transport.write(
            b"".join(pamqp.frame.marshal(frame, self._channel_id) for frame in frames)
        )



from typing import Iterator, List


class ChannelIdStore:
    """Place to organize channel IDs.

    Callers should acquire() from here to find an unused channel ID. They
    should release() any channel ID that is no longer used, so that acquire()
    doesn't run out of options.
    """

    def __init__(self, channel_max: int):
        self._new: Iterator[int] = iter(range(1, channel_max))
        self._recycled: List[int] = []

    def acquire(self) -> int:
        if self._recycled:
            return self._recycled.pop()
        else:
            return next(self._new)

    def release(self, channel_id: int) -> None:
        self._recycled.append(channel_id)



from __future__ import annotations

import asyncio
import ssl
from functools import partial
from typing import Optional, Union, cast
from urllib.parse import urlparse

import pamqp.commands
import pamqp.common

from ._consume_channel import ConsumeChannel
from ._protocol import Protocol


class Connection:
    def __init__(
        self,
        url: str,
        *,
        ssl: Union[ssl.SSLContext, None] = None,
        connect_timeout: Optional[float] = None,
    ):
        self._url = url
        self._ssl = ssl
        self._connect_timeout = connect_timeout

    async def connect(self) -> None:
        """Alternative to `async with` syntax.

        Return if we are connected. Raise if connect fails.

        Do not call this twice on one connection.
        """
        url = urlparse(self._url)
        if url.scheme not in {"amqp", "amqps"}:
            raise ValueError(
                "url must start with 'amqp://' or 'amqps://'; got: %r" % self._url
            )

        if self._ssl is not None and url.scheme == "amqp":
            raise ValueError(
                "When `url` starts with 'amqp://', you must not provide `ssl`. Either omit the `ssl` argument, or make `url` start with 'amqps://'"
            )

        addr = url.netloc.split("@")[-1].split(":")
        if len(addr) == 1:
            host = addr[0]
            port = {"amqp": 5672, "amqps": 5671}[url.scheme]
        else:
            host = addr[0]
            port = int(addr[1])

        transport, protocol = await asyncio.wait_for(
            asyncio.get_running_loop().create_connection(
                protocol_factory=partial(
                    Protocol,
                    username=url.username,
                    password=url.password,
                    virtual_host=url.path,
                ),
                host=host,
                port=port,
                ssl=self._ssl or url.scheme == "amqps",
            ),
            timeout=self._connect_timeout,
        )
        self._protocol = cast(Protocol, protocol)
        await self._protocol.open

    async def __aenter__(self):
        await self.connect()
        return self

    @property
    def closed(self) -> asyncio.Future[None]:
        """Future that resolves when we are closed.

        If closing fails, this Future will have an exception.
        """
        return self._protocol.closed

    async def close(self):
        """Alternative to `async with` syntax.

        You may call this multiple times on a single connection.
        """
        self._protocol.send_close_if_allowed()
        await self.closed

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.close()  # or raise an exception atop the original exception

    def acking_consumer(
        self, queue_name: str, *, prefetch_size: int = 0, prefetch_count: int = 1
    ) -> ConsumeChannel:
        """Consume messages asynchronously.

        Usage:

            async with connection.acking_consumer("my-queue") as consumer:
                async for message in consumer:
                    print("hi! %r" % message)

        After the inner block completes, the consumer will ack the message.

        If your inner block raises an exception, the message will _not_ be
        acked. Once the outer block exits, RabbitMQ may deliver the message
        to another consumer.

        You may call `consumer.close()` to stop the flow of messages in the
        `async for` loop. You probably shouldn't call it from _within_ the
        `async for` loop because a low-traffic queue won't yield to your code
        in a timely manner.

        `prefetch_size` and `prefetch_count` hint to RabbitMQ how much of a
        backlog should be on this queue. Smaller numbers are "fairer" because a
        slow consumer won't hold on to many messages. Larger numbers are
        "faster" because messages will be available locally before your acks
        reach the server.
        """
        return self._protocol.acking_consumer(
            queue_name=queue_name,
            prefetch_size=prefetch_size,
            prefetch_count=prefetch_count,
        )

    def publish(
        self, message: bytes, *, exchange_name: str = "", routing_key: str = ""
    ) -> asyncio.Future[None]:
        return self._protocol.publish(
            message, exchange_name=exchange_name, routing_key=routing_key
        )

    def exchange_declare(
        self,
        exchange_name: str,
        *,
        exchange_type: str = "direct",
        passive: bool = False,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        arguments: pamqp.common.Arguments = None,
    ) -> asyncio.Future[None]:
        return self._protocol.rpc(
            pamqp.commands.Exchange.Declare(
                exchange=exchange_name,
                exchange_type=exchange_type,
                passive=passive,
                durable=durable,
                auto_delete=auto_delete,
                internal=internal,
                nowait=True,
                arguments=arguments,
            )
        )

    def queue_bind(
        self,
        queue_name: str,
        exchange_name: str,
        *,
        routing_key: str = "",
        arguments: pamqp.common.Arguments = None,
    ) -> asyncio.Future[None]:
        return self._protocol.rpc(
            pamqp.commands.Queue.Bind(
                queue=queue_name,
                exchange=exchange_name,
                routing_key=routing_key,
                nowait=True,
                arguments=arguments,
            )
        )

    def queue_declare(
        self,
        queue_name: str,
        *,
        passive: bool = False,
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        arguments: pamqp.common.Arguments = None,
    ) -> asyncio.Future[None]:
        return self._protocol.rpc(
            pamqp.commands.Queue.Declare(
                queue=queue_name,
                passive=passive,
                durable=durable,
                exclusive=exclusive,
                auto_delete=auto_delete,
                nowait=True,
                arguments=arguments,
            )
        )

    def queue_unbind(
        self,
        queue_name: str,
        exchange_name: str,
        *,
        routing_key: str = "",
        arguments: pamqp.common.Arguments = None,
    ) -> asyncio.Future[None]:
        return self._protocol.rpc(
            pamqp.commands.Queue.Unbind(
                queue=queue_name,
                exchange=exchange_name,
                routing_key=routing_key,
                arguments=arguments,
            ),
            ignored_frame_types=frozenset([pamqp.commands.Queue.UnbindOk]),
        )


def connect(
    url: str,
    *,
    ssl: Union[ssl.SSLContext, None] = None,
    connect_timeout: Optional[float] = None,
) -> Connection:
    return Connection(url, ssl=ssl, connect_timeout=connect_timeout)



class BadBytesFromServer(Exception):
    """The server did not give a correct frame"""


class ConnectionClosedByHeartbeatMonitor(Exception):
    """The server did not send any data in too long."""


class ConnectionClosedByServer(Exception):
    """The server sent a Close() frame."""

    def __init__(self, reply_code: int, reply_text: str) -> None:
        super().__init__(self, reply_code, reply_text)
        self.reply_code = reply_code
        self.reply_text = reply_text

    def __str__(self):
        return "RabbitMQ closed the connection: %d %s" % (
            self.reply_code,
            self.reply_text,
        )


class ConnectionClosed(Exception):
    """The connection is closed, so your operation cannot go on."""


class ServerSentNack(Exception):
    """The RabbitMQ sent a "nack" in publisher_confirms mode."""


class ChannelClosed(Exception):
    """The channel is closed, so your operation cannot go on."""


class ChannelClosedByServer(Exception):
    """The server sent a Close() frame."""

    def __init__(self, reply_code: int, reply_text: str) -> None:
        super().__init__(self, reply_code, reply_text)
        self.reply_code = reply_code
        self.reply_text = reply_text

    def __str__(self):
        return "RabbitMQ closed the channel: %d %s" % (
            self.reply_code,
            self.reply_text,
        )



from __future__ import annotations

import asyncio
import logging
from functools import singledispatchmethod
from typing import Callable, List, NamedTuple, Optional, Tuple

import pamqp.base
import pamqp.body
import pamqp.commands
import pamqp.exceptions
import pamqp.header

from ._channel import AcceptableFrame, Channel
from ._exceptions import ChannelClosed, ChannelClosedByServer, ConnectionClosed
from ._frame_writer import FrameWriter

logger = logging.getLogger(__name__)


class _Delivery(NamedTuple):
    deliver: pamqp.commands.Basic.Deliver
    header: Optional[pamqp.header.ContentHeader] = None
    body: List[pamqp.body.ContentBody] = []

    def add_HEADER(self, header: pamqp.header.ContentHeader) -> _Delivery:
        """Return this _Delivery with its header frame."""
        return self._replace(header=header)

    def add_BODY(self, body: pamqp.body.ContentBody) -> _Delivery:
        """Return a _Delivery with an added body frame."""
        return self._replace(body=self.body + [body])


async def _next_delivery(
    queue: asyncio.Queue, closed: asyncio.Future
) -> Tuple[bytes, int]:
    delivery_task = asyncio.create_task(queue.get())
    done, pending = await asyncio.wait(
        (delivery_task, closed), return_when=asyncio.FIRST_COMPLETED
    )
    if delivery_task in done:
        delivery = delivery_task.result()
        return (
            b"".join(body.value for body in delivery.body),
            delivery.deliver.delivery_tag,
        )
    else:  # `self.closed` in Done
        delivery_task.cancel()
        closed.result()  # raise exception if there is one
        raise ChannelClosed


class ConsumeChannelIterator:
    def __init__(
        self,
        queue: asyncio.Queue[_Delivery],
        closed: asyncio.Future[None],
        ack: Callable[[int], None],
    ):
        self._queue = queue
        self._closed = closed
        self._ack = ack
        self._yielded_delivery_tag: Optional[int] = None

    def __aiter__(self):
        return self  # pragma: no cover

    async def __anext__(self):
        if self._yielded_delivery_tag is not None:
            self._ack(self._yielded_delivery_tag)

        try:
            message, self._yielded_delivery_tag = await _next_delivery(
                self._queue, self._closed
            )
            return message
        except ChannelClosed:
            raise StopAsyncIteration from None
        # Other exceptions we'll re-raise.


class ConsumeChannel(Channel):
    """A Channel that presents messages as an async iterator."""

    def __init__(
        self,
        frame_writer: FrameWriter,
        *,
        queue_name: str,
        prefetch_size: int,
        prefetch_count: int,
    ):
        self._frame_writer = frame_writer
        self._delivery: Optional[_Delivery] = None
        self._queue: asyncio.Queue[_Delivery] = asyncio.Queue()
        self.closed: asyncio.Future[None] = asyncio.Future()

        self._frame_writer.send_frames(
            [
                pamqp.commands.Channel.Open(),
                pamqp.commands.Basic.Qos(
                    prefetch_size=prefetch_size, prefetch_count=prefetch_count
                ),
                pamqp.commands.Basic.Consume(
                    queue=queue_name, consumer_tag="carehare", nowait=True
                ),
            ]
        )

    def ack(self, delivery_tag: int):
        """Write Basic.Ack to RabbitMQ.

        This method is _synchronous_: it marks the _beginning_ of an ack. You
        can't know whether RabbitMQ will receive the ack.

        If you call this during channel close or connection close, nothing will
        happen.
        """
        self._frame_writer.send_frame(
            pamqp.commands.Basic.Ack(delivery_tag=delivery_tag),
        )

    @singledispatchmethod
    def accept_frame(self, frame: AcceptableFrame) -> None:
        raise NotImplementedError("Unexpected frame: %r" % frame)  # pragma: no cover

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Channel.OpenOk) -> None:  # type: ignore[no-redef]
        pass

    def _raise_if_delivery_is_set(self):
        if self._delivery:
            # ref: 4.2.6
            # Note that any non-content frame explicitly marks the end of the
            # content. Although the size of the content is well-known from the
            # content header (and thus also the number of content frames), this
            # allows for a sender to abort the sending of content without the
            # need to close the channel.
            #
            # Guidelines for implementers:
            #  A peer that receives an incomplete or badly-formatted content
            # MUST raise a connection exception with reply code 505 (unexpected
            # frame). This includes missing content headers, wrong class IDs in
            # content headers, missing content body frames, etc.
            #
            # ... [adamhooper, 2021-02-15] these two paragraphs seem to
            # conflict; isn't a 'connection exception' a hard error? Let's call
            # it a hard error.
            raise pamqp.exceptions.AMQPUnexpectedFrame(
                "RabbitMQ server did not finish sending body"
            )

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Basic.QosOk) -> None:  # type: ignore[no-redef]
        self._raise_if_delivery_is_set()

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.commands.Basic.Deliver) -> None:  # type: ignore[no-redef]
        self._raise_if_delivery_is_set()
        self._delivery = _Delivery(frame)

    def _enqueue_delivery_if_complete(self):
        received = sum(len(frame.value) for frame in self._delivery.body)
        if received >= self._delivery.header.body_size:
            self._queue.put_nowait(self._delivery)
            self._delivery = None

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.header.ContentHeader) -> None:  # type: ignore[no-redef]
        assert self._delivery is not None
        assert self._delivery.header is None
        self._delivery = self._delivery.add_HEADER(frame)  # Raises AMQPUnexpectedFrame
        self._enqueue_delivery_if_complete()

    @accept_frame.register  # type: ignore[no-redef]
    def _(self, frame: pamqp.body.ContentBody) -> None:  # type: ignore[no-redef]
        assert self._delivery is not None
        assert self._delivery.header is not None
        self._delivery = self._delivery.add_BODY(frame)
        self._enqueue_delivery_if_complete()

    async def next_delivery(self) -> Tuple[bytes, int]:
        """Receive a (message, delivery_tag) tuple from RabbitMQ.

        The caller must call consumer.ack(delivery_tag), or RabbitMQ may deliver
        the message to another client.

        Raise ConnectionClosed if we (or RabbitMQ) closed the connection.

        Raise ChannelClosedByServer if RabbitMQ sends us an error.

        Raise ChannelClosed if we closed the channel.
        """
        # raise ConnectionClosed, ChannelClosedByServer
        return await _next_delivery(self._queue, self.closed)

    def close(self):
        if not self.closed.done():
            self._frame_writer.send_close()
            # RabbitMQ will send Channel.Close, and then we'll resolve self.closed

    def accept_channel_closed_by_server(
        self, frame: pamqp.commands.Channel.Close
    ) -> None:
        self.closed.set_exception(
            ChannelClosedByServer(frame.reply_code or 0, frame.reply_text)
        )

    def accept_channel_closed_by_us(self) -> None:
        self.closed.set_result(None)

    def accept_connection_closed(self, exc: ConnectionClosed) -> None:
        self.closed.set_exception(exc)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if exc_type:
            logger.exception("Closing consumer")
        self.close()
        await self.closed

    def __aiter__(self):
        return ConsumeChannelIterator(self._queue, self.closed, self.ack)



import asyncio
from functools import singledispatchmethod
from typing import FrozenSet

import pamqp.base
import pamqp.commands
import pamqp.exceptions

from ._channel import AcceptableFrame, Channel
from ._exceptions import ChannelClosedByServer, ConnectionClosed
from ._frame_writer import FrameWriter


class RpcChannel(Channel):
    def __init__(
        self,
        frame: pamqp.base.Frame,
        frame_writer: FrameWriter,
        ignored_frame_types: FrozenSet[type],
    ) -> None:
        self._frame_writer = frame_writer
        self._ignored_frame_types = ignored_frame_types
        self.done: asyncio.Future[None] = asyncio.Future()

        self._frame_writer.send_frames(
            [
                pamqp.commands.Channel.Open(),
                frame,
                pamqp.commands.Channel.Close(
                    reply_code=0,
                    reply_text="",
                    class_id=0,
                    method_id=0,
                ),
            ]
        )

    @singledispatchmethod
    def accept_frame(self, frame: AcceptableFrame) -> None:
        if type(frame) in self._ignored_frame_types:
            return

        raise NotImplementedError("Unexpected frame: %r" % frame)  # pragma: no cover

    @accept_frame.register
    def _(self, frame: pamqp.commands.Channel.OpenOk) -> None:
        pass

    def accept_channel_closed_by_server(
        self, frame: pamqp.commands.Channel.Close
    ) -> None:
        self.done.set_exception(
            ChannelClosedByServer(frame.reply_code or 0, frame.reply_text)
        )

    def accept_channel_closed_by_us(self) -> None:
        self.done.set_result(None)

    def accept_connection_closed(self, exc: ConnectionClosed) -> None:
        self.done.set_exception(exc)



import asyncio
import logging
from typing import Callable

import pamqp.heartbeat

logger = logging.getLogger(__name__)

_HEARTBEAT_BYTES = pamqp.heartbeat.Heartbeat.marshal()

# The docs at https://www.rabbitmq.com/heartbeats.html are _really_
# confusing, because they bring up the term "heartbeat interval" which
# is used nowhere other than in the heartbeats.html file itself -- _and_
# all the RabbitMQ discussion-forum questions and issue reports that
# relate to the term in heartbeats.html.
#
# Here's the gist:
# * The server will send 2 packets per `heartbeat` seconds.
# * We, when receiving, will check every `heartbeat` seconds whether there
#   has been traffic in the last `heartbeat` seconds. If not, we die.
#   (This is what the RabbitMQ server does.)


class HeartbeatSender:
    """Periodically send heartbeat frames over `transport`.

    Calling convention:

    The caller should instantiate HeartbeatSender; it will then write to
    `transport` periodically. (This is incompatible with asyncio's "write
    flow control".)

    Call `cancel()` to clean up.
    """

    def __init__(self, transport: asyncio.WriteTransport, heartbeat: int):
        self._transport = transport
        self._heartbeat = heartbeat
        self._loop = asyncio.get_running_loop()
        self._t0 = self._loop.time()
        self._n = 1
        self._interval = heartbeat / 2
        logger.debug("Sending heartbeats every %fs...", self._interval)
        self._timer = self._loop.call_at(self._t0 + self._interval / 2, self._beat)

    def _beat(self):
        """Must be called by self._timer."""
        logger.debug("Sending heartbeat")
        self._transport.write(_HEARTBEAT_BYTES)
        self._n += 1
        self._timer = self._loop.call_at(
            self._t0 + self._interval * self._n, self._beat
        )

    def cancel(self):
        self._timer.cancel()
        del self._timer


class HeartbeatMonitor:
    """Call `on_dead()` if we have not received a frame in too long.

    From the spec:

        4.2.7 Heartbeat Frames:

        If a peer detects no incoming traffic (i.e. received octets) for two
        heartbeat intervals or longer, it should close the connection without
        following the Connection.Close/Close-Ok handshaking, and log an error.

    Calling convention:

    The Protocol should start a HeartbeatMonitor and call `monitor.reset()`
    whenever it receives a packet. The HeartBeatMonitor will call `on_dead()`
    if the connection is dropped.

    Call `cancel()` to clean up. (This will _not_ call `on_dead()`.)
    """

    def __init__(self, heartbeat: int, on_dead: Callable[[], None]):
        self._heartbeat = heartbeat
        self._on_dead = on_dead
        self._loop = asyncio.get_running_loop()
        self._last_data_time = self._loop.time()
        self._timer = self._loop.call_at(self._last_data_time + heartbeat, self._beat)

    def _beat(self):
        """Must be called by self._timer."""
        time = self._loop.time()
        diff = time - self._last_data_time
        logger.debug(
            "Heartbeat monitor: last data came %fs ago (heartbeat is %ds)",
            diff,
            self._heartbeat,
        )
        if diff > self._heartbeat:
            self._on_dead()
        else:
            self._timer = self._loop.call_later(self._heartbeat, self._beat)

    def reset(self):
        self._last_data_time = self._loop.time()

    def cancel(self):
        self._timer.cancel()
        del self._timer



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['carehare']

package_data = \
{'': ['*']}

install_requires = \
['pamqp>=3.0,<4.0']

setup_kwargs = {
    'name': 'carehare',
    'version': '0.0.10',
    'description': 'asyncio RabbitMQ client',
    'long_description': 'carehare\n========\n\nAsyncio RabbitMQ client that handles all the edge cases.\n\nInstallation\n------------\n\n``pip install carehare``\n\nUsage\n-----\n\nConsumer (``async for``)::\n\n    try:\n        async with carehare.connect("amqps://guest:guest@localhost/") as connection:\n            try:\n                await connection.queue_declare("my-queue", exclusive=True)\n                async with connection.acking_consumer("my-queue") as consumer:\n                    async for message in consumer:\n                        print(repr(message))\n                        # ... if we raise an exception here, we won\'t ack.\n                        #\n                        # ... if we `break` from this loop, we won\'t ack.\n                        #\n                        # Call `consumer.close()` before (or instead of) `break`\n                        # to abort iteration.\n            except carehare.ChannelClosedByServer:\n                logger.info("RabbitMQ told this one consumer to go away")\n            except carehare.ConnectionClosed:\n                # Either RabbitMQ is telling us an error (and the outer context\n                # manager will throw it), or we called connection.close()\n                # ourselves (so we want to close).\n                pass\n    except carehare.ConnectionClosedByServer:\n        # str(error) will give the RabbitMQ error message\n        logger.error("RabbitMQ closed our connection")\n    except carehare.ConnectionClosedByHeartbeatMonitor:\n        logger.error("RabbitMQ went away")\n\nConsumer (``next_delivery``)::\n\n    try:\n        async with carehare.connect("amqps://guest:guest@localhost/") as connection:\n            try:\n                await connection.queue_declare("my-queue", exclusive=True)\n                async with connection.acking_consumer("my-queue") as consumer:\n                    while True:\n                        message, delivery_tag = await consumer.next_delivery()\n                        # You must ack() (with no await). If RabbitMQ doesn\'t\n                        # receive this ack, it may deliver the same message to\n                        # another client.\n                        consumer.ack(delivery_tag)\n                        if message.startswith(b"okay, go away now"):\n                            break\n            except carehare.ChannelClosedByServer:\n                logger.info("RabbitMQ told this one consumer to go away")\n            except carehare.ConnectionClosed:\n                # Either RabbitMQ is telling us an error (and the outer context\n                # manager will throw it), or we called connection.close()\n                # ourselves (so we want to close).\n                pass\n    except carehare.ConnectionClosedByServer:\n        # str(error) will give the RabbitMQ error message\n        logger.error("RabbitMQ closed our connection")\n    except carehare.ConnectionClosedByHeartbeatMonitor:\n        logger.error("RabbitMQ went away")\n\nPublisher::\n\n    try:\n        async with carehare.connect("amqps://guest:guest@localhost/") as connection:\n            try:\n                await connection.publish(b"Hello, world!", routing_key="my-queue")\n            except carehare.ServerSentNack:\n                logger.warn("Failed to publish message")\n            except carehare.ChannelClosedByServer:\n                # str(err) will give the RabbitMQ error message -- for instance,\n                # "404 NOT_FOUND" if the exchange does not exist\n                logger.error("Problem with the exchange")\n    except carehare.ConnectionClosedByServer:\n        # str(error) will give the RabbitMQ error message\n        logger.error("RabbitMQ closed our connection")\n    except carehare.ConnectionClosedByHeartbeatMonitor:\n        logger.error("RabbitMQ went away")\n\nDesign decisions\n----------------\n\n``carehare`` is designed to turn RabbitMQ\'s asynchronous error system into\n_understandable_ Python exceptions.\n\nChannels\n~~~~~~~~\n\nCarehare doesn\'t let you control RabbitMQ Channels. They aren\'t Pythonic. (In\nRabbitMQ, an exception on a channel closes the channel -- and cancels all its\npending operations.)\n\nInstead, carehare uses channels to handle errors. For instance, Queue.Declare\ncosts three operations: Channel.Open, Queue.Declare, Channel.Close. Since the\noperation has its own channel, it won\'t interfere with other operations if it\ncauses an exception.\n\nThere\'s a speed-up for publishing: we lazily open a Channel per *exchange*.\nError codes like "not found", "access refused" and "not implemented" will make\ncarehare raise an exception on all pending publishes on the same exchange. Don\'t\nworry: a normal "Nack" ("message wasn\'t delivered") will only make your single\nmessage fail.\n\nExceptions\n~~~~~~~~~~\n\n"Exceptions" are hardly exceptional: as a programmer, they are your job. These\nones are designed to help you solve them.\n\nConnection methods return ``asyncio.Future`` objects. You must await each one\nand handle its errors.\n\nEven though you\'re using Python async context managers, exceptions can\'t\nhappen *everywhere*. Carehare will only raise when you ``await`` a response\nfrom RabbitMQ. \n\nIn particular, ``consumer.ack()`` will never raise! You must call it from the\nmain event loop, but you won\'t await it.\n\nTo code safely, catch these exceptions religiously:\n\n* ``carehare.ChannelClosedByServer``: RabbitMQ did not like the command you\n  just ran. Read the exception message for details. After you receive this\n  message, you may continue using the RabbitMQ connection.\n* ``carehare.ConnectionClosed``: When the connection shuts down, every pending\n  ``Future`` will raise this. Only the actual ``Connection`` context manager\n  will raise the underlying exception: a ``carehare.ConnectionClosedByServer``\n  with a descriptive error message.\n\nCarehare won\'t raise ``asyncio.Cancelled``.\n\nBack-pressure\n~~~~~~~~~~~~~\n\nThe core logic is synchronous. It\'s simpler to reason about. The downside:\nneither RabbitMQ nor users will wait for buffers to empty before sending more\ndata.\n\nUse application-level logic to make sure you don\'t run out of memory:\n\n* Consuming? Don\'t worry. Use ``prefetch_count`` to limit the number of messages\n  RabbitMQ sends. Always ack: carehare won\'t permit ``no-ack``.\n* Publishing? Carehare forces "publisher confirms", so each publish returns a\n  ``Future``. Your application is responsible for not calling publish() too many\n  times simultaneously. Use an ``asyncio.Semaphore`` or reason about your\n  specific use (for instance, "my server will host max 100 clients, and each\n  client can only publish one message at a time").\n\nComparison to other async RabbitMQ clients\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nOther clients tend to try and achieve "RabbitMQ in Python". They fail when it\ncomes to exceptions:\n\n* ``aiormq``: If your code generates an error during consume, aiormq (4.1.1)\n  will catch it and ignore it -- stalling your program.\n* ``aioamqp``: If your connection produces an unexpected error, aioamqp will\n  catch it and ignore it -- stalling your program. Also, the latest release was\n  in 2019.\n\nThis author believes it\'s too confusing to model RabbitMQ\'s API in Python.\nInstead, carehare models your *intent* in Python.\n\nDependencies\n------------\n\nYou\'ll need Python 3.8+ and a RabbitMQ server.\n\nIf you have Docker, here\'s how to start a development server::\n\n    test-server/prepare-certs.sh  # Create SSL certificates used in tests\n    docker run --rm -it \\\n         -p 5671:5671 \\\n         -p 5672:5672 \\\n         -p 15672:15672 \\\n         -v "/$(pwd)"/test-server:/test-server \\\n         -e RABBITMQ_SSL_CACERTFILE=/test-server/ca.cert \\\n         -e RABBITMQ_SSL_CERTFILE=/test-server/server.cert \\\n         -e RABBITMQ_SSL_KEYFILE=/test-server/server.key \\\n         -e RABBITMQ_SSL_VERIFY=verify_peer \\\n         -e RABBITMQ_SSL_FAIL_IF_NO_PEER_CERT=true \\\n         -e RABBITMQ_CONFIG_FILE=/test-server/rabbitmq \\\n         rabbitmq:3.8.11-management-alpine\n\nDuring testing, see the RabbitMQ management interface at http://localhost:15672.\n\nContributing\n------------\n\nTo add features and fix bugs\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\nFirst, start a development RabbitMQ server (see above).\n\nNow take on the development cycle:\n\n#. ``tox`` # to ensure tests pass.\n#. Write new tests in ``tests/`` and make sure they fail.\n#. Write new code in ``carehare/`` to make the tests pass.\n#. Submit a pull request.\n\nTo deploy\n~~~~~~~~~\n\nUse `semver <https://semver.org/>`_.\n\n#. ``git push`` and make sure Travis tests all pass.\n#. ``git tag vX.X.X``\n#. ``git push --tags``\n\nTravisCI will push to PyPi.\n',
    'author': 'Adam Hooper',
    'author_email': 'adam@adamhooper.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from ._connection import Connection, connect
from ._exceptions import (
    BadBytesFromServer,
    ChannelClosed,
    ChannelClosedByServer,
    ConnectionClosed,
    ConnectionClosedByHeartbeatMonitor,
    ConnectionClosedByServer,
    ServerSentNack,
)

__all__ = [
    "BadBytesFromServer",
    "ChannelClosed",
    "ChannelClosedByServer",
    "Connection",
    "ConnectionClosed",
    "ConnectionClosedByHeartbeatMonitor",
    "ConnectionClosedByServer",
    "ServerSentNack",
    "connect",
]
